/**
 * Cat5171 class
 *
 * Driver class for the CAT5171 digital potentiometer.
 *
 * Created on: May 15, 2020
 * @author: Aric Beaver
 *
 * @copyright Copyright 2020 ON Semiconductor
 *
 * @file Cat5171.cpp
 */

#include <stdint.h>
#include <errno.h>
#include <vector>
#include "mbed.h"
#include "cat5171.h"


/** 
 * @brief Cat5171 class constructor.
 * @param i2c Pointer to mbed i2c class.
 * @param i2c_address The I2C address of the CAT5171 using 8-bit addressing.
 * @param resistance Enum that specifies the device's resistance value. See cat5171.h for valid options.
 */
Cat5171::Cat5171(I2C *i2c, uint8_t i2c_address, Resistance resistance) {
    m_i2c = i2c;
    m_i2c_address = i2c_address;
    m_resistance = resistance;
}

Cat5171::~Cat5171() {
}

// ================ Standard Read Write I2C Functions ================ //

/**
 * @brief Read data from specified I2C register address.
 * @param data Char array of data to be read.
 */
bool Cat5171::readI2C(char *data) {
    return (!m_i2c->read(m_i2c_address, data, CAT5171_DATA_SIZE, false)); // Return true = 1 (ACK), false = 0 (NACK) 
}
    
/**
 * @brief Write data to specified I2C register address.
 * @param ib Instruction byte desired to be written.
 * @param data Char array of data to be written.
 */
bool Cat5171::writeI2C(char ib, char *data) {
    std::vector<char> cmd;
    cmd.push_back(ib);
    cmd.push_back(data[0]);
    return (!m_i2c->write(m_i2c_address, &cmd[0], CAT5171_DATA_SIZE+1 , false)); // Return true = 1 (ACK), false = 0 (NACK) 
}

/**
 * @brief Writes the resistance between digital POT pins W and B. Will return false if resistance value is out of range.
 * @param resistance Resistance value to write.
 */
bool Cat5171::write(uint resistance) {

    if (resistance <= m_resistance) {
        char ib = CAT5171_SD(shutdown);
        char data[CAT5171_DATA_SIZE] = {((float)255*((float)resistance/(float)m_resistance))};
        writeI2C(ib, data);
    } else {
        // Resistance out of range
        return false;
    }
    return true;    
}

/**
 * @brief Reads the resistance between digital POT pins W and B.
 * @param resistance Resistance value uint pointer.
 */
bool Cat5171::read(uint* resistance) {
    char data[CAT5171_DATA_SIZE];
    bool success = readI2C(data);
    *resistance = (uint8_t)*data * (float)(m_resistance/255);
    return success;    
}

/**
 * @brief Moves the wiper to the midscale tap.
 */
bool Cat5171::rs() {
    char ib = CAT5171_RS(true);
    char data[CAT5171_DATA_SIZE] = {0x00}; // Don't care can write any resistance
    return writeI2C(ib, data);
}

/**
 * @brief Shutdown creates an open circuit at terminal A and shorts the wiper terminal W to terminal B.
 * @param value Shutdown value, true = shutdown, false = normal and returns to previous wiper state
 */
bool Cat5171::sd(bool value) {
    if (value != shutdown) {
        shutdown = value;
        char ib = CAT5171_SD(value);
        
        // Need to read resistance to avoid overwriting
        char data[CAT5171_DATA_SIZE];
        bool success = readI2C(data);

        return writeI2C(ib, data);
    }
    return true;
}