/**
 * nct375.h
 *
 * @created: June 10, 2019
 * @author: Nathan Nichols
 * @brief Driver class for the NCT375 Temperature sensor 
 */

#ifndef NCT375_H
#define NCT375_H

#include "mbed.h"
#include <math.h>

class Nct375 {

    public:
        /** Nct375 constructor
         *  @brief Nct375::Nct375
         *  @param i2c object
         *  @param i2c_address using 8-bit addressing
         */
        Nct375(I2C *i2c, uint8_t i2c_address);

        /**
         * @brief Nct375 destructor
         */
        ~Nct375();

        /**
         * @brief Read both temperature registers, decimal precision of .0625
         */
        double readTemperature();

        /**
         * @brief Read integer temperature register, integer precision
         */
        int quickReadTemperature();

        /**
         * @brief Read temperature when in one shot mode, decimal precision of .0625
         *
         * @note: Forces NCT375 into one-shot mode for one conversion if not currently enabled 
         *        and returns back to normal read mode after conversion is done
         */
        double readOneShot();

        /**
         * @brief Read integer temperature register when in oneshot mode, integer precision
         */
        int quickReadOneShot();

        /**
         * @brief Sets the hysteresis value of the sensor.
         * @param double precision of .0625
         */
        bool setTemperatureHyst(double hystTemp);

        /**
         * @brief Read the current hysteresis value, defaults to 75
         */
        double getTemperatureHyst();

        /**
         * @brief Sets the temperature limit value of the sensor.
         * @param double precision of .0625
         */
        bool setTemperatureOs(double osTemp);

        /**
         * @brief Read the current temperature OS/AlertValue, defaults to 80
         */
        double getTemperatureOs();

        /**
         * @brief Turns on or off shutdown mode (defaults to off)
         * @param Bool denoting whether shutdown mode is on(true) or off(false).
         */
        bool setShutdownMode(bool shutdown);

        /**
         * @brief Turns on inturrupt mode (defaults to comparator mode)
         * @param Bool denoting whether inturrupt mode is on(true) or off(false).
         */
        bool setInterruptMode(bool interrupt);

        /**
         * @brief Sets OsAlert polarity (defaults to 0-active low)
         * @param Bool denoting polarity after OS trips (0-active low, 1-active high)
         */
        bool setOsAlertPolarity(bool high);

        /**
         * @brief Sets One Shot Mode (defaults to off)
         * @param Bool denoting on(true) or off(false)
         */
        bool setOneShotMode(bool oneshot);

        /**
         * @brief Sets number of faults to trigger OS/Alert (defaults to 1)
         * @param int denoting number of faults, 1, 2, 4, or 6
         */
        bool setNumFaults(int numFaults);

        /**
         * @brief Read the current configuration register value, defaults to 0x00
         */
        char getConfig();
        
        /**
         * @brief clear interrupt only when in interrupt mode
         */
        bool clearInterrupt();

        //I2C write size based on register size
        static constexpr int NCT375_REGISTER_WRITE_ADDR = 1;
        static constexpr int NCT375_REGISTER_WRITE_8_BIT = 2;
        static constexpr int NCT375_REGISTER_WRITE_16_BIT = 3;

        //I2C read size based on register size
        static constexpr int NCT375_REGISTER_READ_ADDR = 1;
        static constexpr int NCT375_REGISTER_READ_8_BIT = 2;
        static constexpr int NCT375_REGISTER_READ_16_BIT = 3;

        //Register addresses
        enum { TEMPERATURE_ADDR = 0, CONFIG_ADDR = 1, THYST_ADDR = 2, TOS_ADDR = 3, ONE_SHOT_ADDR = 4 };

        //Number of faults before OS/Alert trips
        enum class FaultQueue { ONE_FAULT = 0, TWO_FAULTS = 8, FOUR_FAULTS = 16, SIX_FAULTS = 24};

        //States for oneshot
        enum OneShotStates { READ, ONESHOT, UPDATE };

        //Ints defining indices for MSBs and LSBs of 8 bit values
        static constexpr unsigned int MSBs = 0;
        static constexpr unsigned int LSBs = 1;

        //Constants for "shifting" 
        static constexpr unsigned int DOUBLE_SHIFT_4_BITS = 16;
        static constexpr unsigned int SHIFT_4_BITS = 4;

        //Defaults OS/hysteresis/config values
        static constexpr char DEFAULT_CONFIG = 0x00;
        static constexpr int DEFAULT_HYST = 75;
        static constexpr int DEFAULT_OS = 80;

        //Config data bit
        static constexpr char SHUTDOWN_MODE = 0x01;
        static constexpr char COMP_INT_MODE = 0x02;
        static constexpr char OS_ALERT_POLARITY = 0x04;
        static constexpr char CLEAR_FAULT_QUEUE = 0x18;
        static constexpr char CLEAR_LSBS = 0xF0;
        static constexpr char ONE_SHOT_MODE = 0x20;

        //Typical temperature conversion time
        static constexpr double CONVERSION_TIME = 48500;

        //Minimum thread wait time
        static constexpr unsigned int MIN_THR_WAIT_TIME = 1;

        //Conversion constants
        static constexpr int NEGATIVE_THRESHOLD = 2047;
        static constexpr int MAX_READING = 4096;
        static constexpr double MAX_GUARANTEED_TEMP = 125.00;
        static constexpr double MIN_GUARANTEED_TEMP = -55.00;

    protected:
        /**
         * @member I2C object pointer
         */
        I2C *i2c_;

        /**
         * @member I2C address of this device
         */
        int i2c_address_;

        /**
         * @member current configuration of the device   
         */
        char configuration_;

        /**
         * @brief writes the Address pointer to the specified address
         */
        void setPointerAddress(char address);

        /**
         * @brief writes current configuration to slave mC
         */
        bool writeConfig();

        /**
         * @brief Takes a temperature in double and turns it into the corresponding integer valued adc reading
         */
        int convertTempToAdcValue(double temp);

        /**
         * @brief Takes an adc value and converts it to a temperature
         */
        int convertRegisterToTempValue(unsigned char* adc);

        /**
         * @brief converts an ADC reading to a temperature reading
         */
        double convertAdcValueToTemp(int adc);

        /**
         * @brief converts the character values read from the register to an ADC reading
         */
        int convertRegisterToAdcValue(unsigned char t0, unsigned char t1);

        /**
         * @member Timeout to wait for oneshot conversion
         */
        Timeout oneshot_timer_;

        /**
         * @brief update the value in the NCT375's oneshot register to initiate oneshot temperature reading
         */
        bool triggerOneShotReading();

        /**
         * @brief update current state to oneshot after oneshot conversion triggered
         */
        void updateOneShotState();

        /**
         * @brief States to track oneshot mode
         */
        OneShotStates current_state_;
        OneShotStates prev_state_;

};
 #endif //NCT375_H