/**
 * nct375 class
 *
 * Driver class for the NCT375 temperature sensor
 *
 * Created on: June 10, 2019
 * @author: Nathan Nichols
 *
 * @copyright Copyright 2019 ON Semiconductor
 *
 * @file nct375.cpp
 */

#include "nct375.h"

Nct375::Nct375(I2C *i2c, uint8_t i2c_address) :
    i2c_(i2c),
    i2c_address_(i2c_address),
    configuration_(DEFAULT_CONFIG),
    current_state_(READ),
    prev_state_(READ)
{

    //Initialize configuration register
    writeConfig();

    //Set Os/Alert threshold and hysteresis to default values
    setTemperatureHyst(DEFAULT_HYST);
    setTemperatureOs(DEFAULT_OS);

}

Nct375::~Nct375() {
}

double Nct375::readTemperature() {
    unsigned char temperature[NCT375_REGISTER_READ_16_BIT];
    setPointerAddress(TEMPERATURE_ADDR);
    i2c_->read(i2c_address_, (char*)temperature, NCT375_REGISTER_READ_16_BIT);
    int adcValue = convertRegisterToAdcValue(temperature[0], temperature[1]);
    return convertAdcValueToTemp(adcValue);
}

int Nct375::quickReadTemperature() {
    char temperature[NCT375_REGISTER_READ_8_BIT];
    setPointerAddress(TEMPERATURE_ADDR);
    i2c_->read(i2c_address_, temperature, NCT375_REGISTER_READ_8_BIT);
    return temperature[MSBs];
}

bool Nct375::setTemperatureHyst(double hystTemp) {
    int adcValue = convertTempToAdcValue(hystTemp);
    char data[NCT375_REGISTER_WRITE_16_BIT];
    data[0] = THYST_ADDR;
    data[1] = adcValue >> SHIFT_4_BITS;
    data[2] = adcValue << SHIFT_4_BITS;
    // Shifting the adc value turns integer (xxxx xxxx yyyy) into: 
    // data[1] = xxxx xxxx
    // data[2] = yyyy 0000
    return !i2c_->write(i2c_address_, data, NCT375_REGISTER_WRITE_16_BIT);
}

double Nct375::getTemperatureHyst() {
    unsigned char temperature[NCT375_REGISTER_READ_16_BIT];
    setPointerAddress(THYST_ADDR);
    i2c_->read(i2c_address_, (char*)temperature, NCT375_REGISTER_READ_16_BIT);
    int adcRead = convertRegisterToAdcValue(temperature[0], temperature[1]);
    return convertAdcValueToTemp(adcRead);
}

bool Nct375::setTemperatureOs(double osTemp) {
    int adcValue = convertTempToAdcValue(osTemp);
    char data[NCT375_REGISTER_WRITE_16_BIT];
    data[0] = TOS_ADDR;
    data[1] = adcValue >> SHIFT_4_BITS;
    data[2] = adcValue << SHIFT_4_BITS;
    // Shifting the adc value turns integer (xxxx xxxx yyyy) into: 
    // data[1] = xxxx xxxx
    // data[2] = yyyy 0000
    return !i2c_->write(i2c_address_, data, NCT375_REGISTER_WRITE_16_BIT);
}

double Nct375::getTemperatureOs() {
    unsigned char temperature[NCT375_REGISTER_READ_16_BIT];
    setPointerAddress(TOS_ADDR);
    i2c_->read(i2c_address_, (char*)temperature, NCT375_REGISTER_READ_16_BIT);
    int adcRead = convertRegisterToAdcValue(temperature[0], temperature[1]);
    return convertAdcValueToTemp(adcRead);
}

bool Nct375::writeConfig() {
    char data[NCT375_REGISTER_WRITE_8_BIT];
    data[0] = CONFIG_ADDR;
    data[1] = configuration_;
    return !i2c_->write(i2c_address_, data, NCT375_REGISTER_WRITE_8_BIT);
}

char Nct375::getConfig() {
    char data[NCT375_REGISTER_READ_8_BIT];
    setPointerAddress(CONFIG_ADDR);
    i2c_->read(i2c_address_, data, NCT375_REGISTER_READ_8_BIT);
    return data[MSBs];
}

bool Nct375::setShutdownMode(bool shutdown) {
    configuration_ &= ~SHUTDOWN_MODE;
    configuration_ |= shutdown ? SHUTDOWN_MODE : 0x00;
    return writeConfig();
}

bool Nct375::setInterruptMode(bool interrupt) {
    configuration_ &= ~COMP_INT_MODE;
    configuration_ |= interrupt ? COMP_INT_MODE : 0x00;
    return writeConfig();
}

bool Nct375::setOsAlertPolarity(bool high) {
    configuration_ &= ~OS_ALERT_POLARITY;
    configuration_ |= high ? OS_ALERT_POLARITY : 0x00;
    return writeConfig();
}

bool Nct375::setOneShotMode(bool oneshot) {
    prev_state_ = current_state_;
    current_state_ = oneshot ? ONESHOT : READ;
    if (current_state_ == prev_state_) {
        return true;
    }
    configuration_ &= ~ONE_SHOT_MODE;
    configuration_ |= oneshot ? ONE_SHOT_MODE : 0x00;
    return writeConfig();
}

bool Nct375::setNumFaults(int numFaults) {
    FaultQueue faultQueueConfig_;
    switch (numFaults) {
        case 1: faultQueueConfig_ = FaultQueue::ONE_FAULT;
            break;
        case 2: faultQueueConfig_ = FaultQueue::TWO_FAULTS;
            break;
        case 4: faultQueueConfig_ = FaultQueue::FOUR_FAULTS;
            break;
        case 6: faultQueueConfig_ = FaultQueue::SIX_FAULTS;
            break;
        default: return false;
    }
    configuration_ &= ~CLEAR_FAULT_QUEUE; 
    configuration_ |= static_cast<int>(faultQueueConfig_);
    return writeConfig();
}

bool Nct375::clearInterrupt() {
    char data[1];
    return !i2c_->read(i2c_address_, data, NCT375_REGISTER_READ_8_BIT);
}

void Nct375::setPointerAddress(char address) {
    i2c_->write(i2c_address_, &address, NCT375_REGISTER_WRITE_ADDR); 
}

bool Nct375::triggerOneShotReading() {

    char data[NCT375_REGISTER_WRITE_8_BIT] = {ONE_SHOT_ADDR, 0x00};
    bool oneshot_started = !i2c_->write(i2c_address_, data, NCT375_REGISTER_WRITE_8_BIT);
    current_state_ = UPDATE;
    oneshot_timer_.attach_us(callback(this, &Nct375::updateOneShotState), CONVERSION_TIME);
    return oneshot_started;
}

void Nct375::updateOneShotState() {
    current_state_ = ONESHOT;
    oneshot_timer_.detach();
}

double Nct375::readOneShot() {

    //Wait if another oneshot conversion is currently active
    while (current_state_ == UPDATE) {
        Thread::wait(MIN_THR_WAIT_TIME);
    }

    //Change to oneshot mode if not currently enabled
    if (current_state_ != ONESHOT) {
        setOneShotMode(true);
    }

    //Start oneshot conversion
    triggerOneShotReading();

    //Wait for triggered conversion to finish
    while (current_state_ == UPDATE){
        Thread::wait(MIN_THR_WAIT_TIME);
    }

    //If previously configured for normal read mode, return to that state
    if (prev_state_ == READ) {
        setOneShotMode(false);
    }

    return readTemperature();
}

int Nct375::quickReadOneShot() {

    //Wait if another oneshot conversion is currently active
    while (current_state_ == UPDATE) {
        Thread::wait(MIN_THR_WAIT_TIME);
    }

    //Change to oneshot mode if not currently enabled
    if (current_state_ != ONESHOT) {
        current_state_ = ONESHOT;
    }

    //Start oneshot conversion
    triggerOneShotReading();

    //Wait for triggered conversion to finish
    while (current_state_ == UPDATE){
        Thread::wait(MIN_THR_WAIT_TIME);
    }

    //If previously configured for normal read mode, return to that state
    if (prev_state_ == READ) {
        current_state_ = READ;
    }
    
    return quickReadTemperature();
}

/**
 * From Documentation
 * Positive Temperature = ADC Code (decimal)/16 Example 190h = 400d/16 = +25°C
 * Negative Temperature = (ADC Code(decimal) - 4096)/16 Example E70h = (3696d - 4096)/16 = -25°C
 * Algebraically solve for ADC Code
 * (Positive) ADC = Temperature * 16
 * (Negative) ADC = (16*Temperature) + 4096
 */
int Nct375::convertTempToAdcValue(double temp) {
    temp = (temp > MAX_GUARANTEED_TEMP) ? MAX_GUARANTEED_TEMP : temp;
    temp = (temp < MIN_GUARANTEED_TEMP) ? MIN_GUARANTEED_TEMP : temp;
 
    return (temp >= 0) ? (int)(temp * DOUBLE_SHIFT_4_BITS) : (int)((DOUBLE_SHIFT_4_BITS * temp) + MAX_READING);
}

/**
 * Register read in the form xxxx xxxx yyyy 0000
 * temp[MSBs] = xxxx xxxx
 * temp[LSBs] = yyyy 0000
 * converts to the integer value: 
 *       xxxx xxxx yyyy
 */
int Nct375::convertRegisterToAdcValue(unsigned char t0, unsigned char t1) {
    return ((int)t0 * DOUBLE_SHIFT_4_BITS) + ((int)t1 / DOUBLE_SHIFT_4_BITS);
}

/**
 * From Documentation
 * Positive Temperature = ADC Code (decimal)/16
 * Example 190h = 400d/16 = +25°C
 * Negative Temperature = (ADC Code(decimal) - 4096)/16 
 * Example E70h = (3696d - 4096)/16 = -25°C
 */
double Nct375::convertAdcValueToTemp(int adc) {
    return (adc > NEGATIVE_THRESHOLD) ? ((double)(adc-MAX_READING))/DOUBLE_SHIFT_4_BITS : ((double)adc)/DOUBLE_SHIFT_4_BITS;
}