/**
 * Ncv768x.h
 *
 * @created: Apr 03, 2020
 * @author: Aric Beaver
 * @brief Driver class for the NCV7684 and NCV7685 LED driver.
 */

#ifndef NCV768X_H
#define NCV768X_H

#include <string>

/* ============== BASE REGISTER IDS ============== */

// Read voliate registers (same between NCV7684 and NCV7685)
// Status
#define NCV768X_ID_STATUS_R             0x08 // I2C_STATUS
#define NCV768X_ID_STATUS_R_SIZE        1
#define NCV768X_ID_STATUS_R_FLAGS       8
// Fault
#define NCV768X_ID_FAULT_R              0x09 // FAULT_STATUS
#define NCV768X_ID_FAULT_R_SIZE         2
// Read all but only using I2C_CH_STATUS
#define NCV768X_ID_READALL_R            0x0A // I2C_CH_STATUS, I2C_STATUS, FAULT_STATUS
#define NCV768X_ID_READALL_R_SIZE       5

// Write volatile registers
// same
#define NCV768X_ID_I2C_CONF_W           0x00 // I2C_CONF
#define NCV768X_ID_I2C_CONF_W_SIZE      2
// NCV7684                                                                      // changed, need to update NCV7684 text
#define NCV768X_ID_PWM_NCV7684_W            0x01 // PWM_GAIN, PWM_GAIN_EN               // ID_PWM_W
#define NCV768X_ID_PWM_NCV7684_W_SIZE       3                                           // ID_PWM_W_SIZE
#define NCV768X_ID_WRITEALL_NCV7684_W       0x02 // I2C_CONF, PWM_GAIN, PWM_GAIN_EN     // ID_WRITEALL_W
#define NCV768X_ID_WRITEALL_NCV7684_W_SIZE  5                                           // ID_WRITEALL_W_SIZE
#define NCV768X_ID_PWM_CONF_NCV7684_W       0x03 // PWM_CONF                            // ID_PWM_CONF_W
#define NCV768X_ID_PWM_CONF_NCV7684_W_SIZE  1                                           // ID_PWM_CONF_W_SIZE
// NCV7685
#define NCV768X_ID_PWM_NCV7685_W            0x01 // PWM_DUTY
#define NCV768X_ID_PWM_NCV7685_W_SIZE       1
#define NCV768X_ID_PWM_CONF_NCV7685_W       0x02 // PWM_CONF, PWM_DUTY_EN
#define NCV768X_ID_PWM_CONF_NCV7685_W_SIZE  3
#define NCV768X_ID_PWM_ALL_NCV7685_W        0x03 // PWM_D1-12
#define NCV768X_ID_PWM_ALL_NCV7685_W_SIZE   12
#define NCV768X_ID_WRITEALL_NCV7685_W       0x04 // I2C_CONF, PWM_CONF, PWM_DUTY_EN
#define NCV768X_ID_WRITEALL_NCV7685_W_SIZE  5

// OTP registers, read and write
#define NCV768X_ID_SET_OTP_W            0x20 // SAM_CONF_1, SAM_CONF_2, ADD_SAM_SET
#define NCV768X_ID_LOCK_OTP_W           0x21 // SAM_CONF_1, SAM_CONF_2, ADD_SAM_SET
#define NCV768X_ID_OTP_W_SIZE           5

#define NCV768X_ID_READ_OTP_R           0x28 // ID_VERS_1, ID_VERS_2, SAM_CONF_1, SAM_CONF_2, ADD_SAM_SET
#define NCV768X_ID_READ_OTP_R_SIZE      7

// CRC constants
#define NCV768X_CRC_POLY 0x2F // x8+x5+x3+x2+x+1 
#define NCV768X_CRC_BITS 8
#define NCV768X_CRC_INIT_XOR 0xFF
#define NCV768X_CRC_FINAL_XOR 0x00
#define NCV768X_CRC_REFLECT_DATA 0
#define NCV768X_CRC_REFLECT_REMAINDER 0

/* ============== CLASS DEFINITION ============== */

class Ncv768x {
public:

    // Part number selection enum
    enum PartNumber {NCV7684 = 0, NCV7685 = 1};

	Ncv768x(I2C *i2c, uint8_t i2c_address, PartNumber part_number, bool crc_en);

	~Ncv768x();

	bool readI2C(char address, uint32_t size, char *data);
	bool writeI2C(char id, uint32_t size, char *data);

    bool i2cStatus(bool i2c_status[NCV768X_ID_STATUS_R_FLAGS]);
    const static uint8_t i2c_status_flags = NCV768X_ID_STATUS_R_FLAGS;
    const string i2c_status_names[NCV768X_ID_STATUS_R_FLAGS] = {"OL", "DIAGERR", "TSD", "TW", "diagRange", "UV", "I2Cerr", "SC_Iset"};

    const static uint8_t ch_size = 12;

    bool faultStatus(bool fault_status[ch_size]);
    
    bool i2cEnable(bool i2c_en);

    bool i2cChEnableRead(bool i2c_ch_enable[ch_size]);
    bool i2cChEnableWrite(bool i2c_ch_enable[ch_size]);
    
    enum DiagMode {OFF = 0, AUTO_RETRY = 1, DIAG_ONLY = 2};

    bool i2cDiagMode(DiagMode diag_mode);    
    bool pwmEnable(bool pwm_en); 

    enum PwmFreq {HZ_125 = 0, HZ_150 = 1, HZ_250 = 2, HZ_300 = 3, HZ_500 = 4, HZ_600 = 6, HZ_1200 = 7};
    bool pwmConf(PwmFreq pwm_freq, bool pwm_lin, uint8_t pwm_duty[ch_size], bool pwm_en[ch_size]); 

    bool readOtp(uint8_t* id_vers_1, uint8_t* id_vers_2, bool sam_conf_1[ch_size], bool sam_conf_2[ch_size], DiagMode* diag_mode, bool* crc_en, uint8_t* bus_addr);

    bool writeOtp(bool otp, bool sam_conf_1[ch_size], bool sam_conf_2[ch_size], DiagMode diag_mode, bool crc_en, uint8_t bus_addr);

    bool checkOtp();
    bool checkI2C();

    bool readI2CCRC();

protected:
    /**
     * @member I2C object pointer
     */
    I2C *m_i2c;

	/**
	 * @member I2C address of this device
	 */
	int m_i2c_address;

    /**
	 * @member Part number of the device configured. Can be NCV7684 or NCV7685.
	 */
    PartNumber m_part_number;

    /**
	 * @member Enables or disables CRC I2C.
	 */
    bool m_crc_en;
    
};
#endif /* Ncv768x_H */
