/**
 * Ncv768x class
 *
 * Driver class for the NCV7684 and NCV7685 LED Driver.
 *
 * Created on: Apr 03, 2020
 * @author: Aric Beaver
 *
 * @copyright Copyright 2020 ON Semiconductor
 *
 * @file Ncv768x.cpp
 */

#include <stdint.h>
#include <errno.h>
#include <vector>
#include "mbed.h"
#include "ncv768x.h"

// CRC definition for I2C
MbedCRC<NCV768X_CRC_POLY, NCV768X_CRC_BITS> m_crc(NCV768X_CRC_INIT_XOR, NCV768X_CRC_FINAL_XOR, NCV768X_CRC_REFLECT_DATA, NCV768X_CRC_REFLECT_REMAINDER);

/** 
 * @brief Ncv768x class constructor.
 * @param i2c Pointer to mbed i2c class.
 * @param i2c_address The I2C address of the Ncv768x using 8-bit addressing.
 * @param part_number Enum that specifies the device part number that is connected. See ncv768x.h for valid options.
 * @param crc_en Enables or disables CRC I2C.
 */
Ncv768x::Ncv768x(I2C *i2c, uint8_t i2c_address, PartNumber part_number, bool crc_en) {
    m_i2c = i2c;
    m_i2c_address = i2c_address;
    m_part_number = part_number;
    m_crc_en = crc_en;
}

Ncv768x::~Ncv768x() {
}

// ================ Standard Read Write I2C Functions ================ //

/**
 * @brief Check for proper I2C communication. Returns true when I2C communication is successful and false otherwise.
 */
bool Ncv768x::checkI2C() {
    // Read fault status registers to ensure ACK was generated on I2C bus.
    bool fault_status[ch_size];
    return faultStatus(fault_status);
}

bool Ncv768x::readI2CCRC() {
    return m_crc_en;
}

/**
 * @brief Read data from specified I2C register address.
 * @param id Address of byte desired to be read.
 * @param size Size of data to be returned for each id.
 * @param data Char array of data read.
 */
bool Ncv768x::readI2C(char id, uint32_t size, char *data) {

    // Write
    bool success = true;
    std::vector<char> cmd;
    cmd.push_back(m_i2c_address);
    cmd.push_back(id);
    if (m_crc_en) {
        // Add initial I2C address at beginning of vector but don't add to cmd as write() already does this
        std::vector<char> cmd_crc = cmd;
        cmd_crc.insert(cmd_crc.begin(), m_i2c_address);
        uint32_t crc_byte = 0;
        m_crc.compute(&cmd_crc[0], cmd_crc.size(), &crc_byte);
        // Add CRC at end of I2C write frame
        cmd.push_back(crc_byte);
    }
    success = success && !m_i2c->write(m_i2c_address, &cmd[0], cmd.size(), true);

    // Read
    if (m_crc_en) {
        // Convert data array to vector to manupulate size
        std::vector<char> data_crc(data, data + size);

        // Write and get CRC, need to add a byte at the end for the CRC byte, then erase before CRC calc
        data_crc.push_back(0x00);
        success = success && !m_i2c->read(m_i2c_address, &data_crc[0], data_crc.size(), false);
        uint32_t crc_byte_read = data_crc[data_crc.size()-1];
        data_crc.pop_back();

        // Calculate CRC by adding I2C address at beginning then removing before returning data
        data_crc.insert(data_crc.begin(), (m_i2c_address | 0b00000001)); // OR with 0b1 for CRC calc 
        uint32_t crc_byte = 0;
        m_crc.compute(&data_crc[0], data_crc.size(), &crc_byte);
        data_crc.erase(data_crc.begin());

        if (success && (crc_byte_read == crc_byte)) {
            // Write success, CRC matched, copy contents of vector to char array
            //data = &data_crc[0];
            std::copy(data_crc.begin(), data_crc.end(), data);
        } else {
            return false;
        }
    } else {
        success = success && !m_i2c->read(m_i2c_address, data, size, false);
    }
    
    return success; // Return true = 1 (ACK), false = 0 (NACK) 
}
    
/**
 * @brief Write data to specified I2C register address.
 * @param id Address of byte desired to be written.
 * @param size Size of data to be written for each id.
 * @param data Char array of data to be written.
 */
bool Ncv768x::writeI2C(char id, uint32_t size, char *data) {
    std::vector<char> cmd;
    cmd.push_back(m_i2c_address);
    cmd.push_back(id);
    for (uint32_t i = 0; i < size; i++) {
        cmd.push_back(data[i]);
    }
    if (m_crc_en) {
        // Add initial I2C address at beginning of vector but don't add to cmd as write() already does this
        std::vector<char> cmd_crc = cmd;
        cmd_crc.insert(cmd_crc.begin(), m_i2c_address);
        uint32_t crc_byte = 0;
        m_crc.compute(&cmd_crc[0], cmd_crc.size(), &crc_byte);
        // Add CRC at end of I2C write frame
        cmd.push_back(crc_byte);
    }
    // &vector[0] pointer indicates the start of the char array
    return !m_i2c->write(m_i2c_address, &cmd[0], cmd.size(), false); // Return true = 1 (ACK), false = 0 (NACK) 
}

// ================ Volatile Read Registers ================ //

/**
 * @brief Reads the status of the I2C_STATUS register.
 * @param i2c_status Boolean array of flags from I2C_STATUS register. The constant flag names are in i2c_status_names variable.
 */
bool Ncv768x::i2cStatus(bool i2c_status[NCV768X_ID_STATUS_R_FLAGS]) {
    char data[NCV768X_ID_STATUS_R_SIZE];
    if(!readI2C(NCV768X_ID_STATUS_R, NCV768X_ID_STATUS_R_SIZE, data)) {
        return false; // I2C NACK
    }
    for (uint8_t i = 0; i < NCV768X_ID_STATUS_R_FLAGS; i++) {
        i2c_status[i] = (data[0] >> i) & 0b1;
    }
    return true;    
}

/**
 * @brief Reads the status of the FAULT_STATUS register.
 * @param fault_status Boolean array containing the fault status of OUT1-12, true = fault/open load, false = normal operation.
 */
bool Ncv768x::faultStatus(bool fault_status[ch_size]) {
    char data[NCV768X_ID_FAULT_R_SIZE];
    if(!readI2C(NCV768X_ID_FAULT_R, NCV768X_ID_FAULT_R_SIZE, data)) {
        return false; // I2C NACK
    }
    char16_t value = (data[0] << 8) | data[1];
    for (uint8_t i = 0; i < ch_size; i++) {
        fault_status[i] = (value >> i) & 0b1;
    }
    return true;    
}

/**
 * @brief Reads the enable status of OUT1-12.
 * @param i2c_ch_enable Boolean array containing the enable status of OUT1-12. The enable status will always read enabled when in PWM mode.
 */
bool Ncv768x::i2cChEnableRead(bool i2c_ch_enable[ch_size]) {
    // On caveat about this read from the NCV7684 datasheet:
    // When NCV7684 is configured in I2C mode and output channel
    // OUTx is configured to operate in PWM mode, I2C_CH_STATUS[x]
    // shall contain value 1.
    char data[NCV768X_ID_READALL_R_SIZE];
    if(!readI2C(NCV768X_ID_READALL_R, NCV768X_ID_READALL_R_SIZE, data)) {
        return false; // I2C NACK
    }
    char16_t value = (data[0] << 8) | data[1];
    for (uint8_t i = 0; i < ch_size; i++) {
        i2c_ch_enable[i] = (value >> i) & 0b1;
    }
    return true;    
}

// ================ Volatile Write Registers ================ //

/**
 * @brief Toggles betweeen I2C and SAM (stand alone mode) modes. 
 * @param i2c_en Boolean to toggle mode, true = I2C mode, false = SAM.
 */
bool Ncv768x::i2cEnable(bool i2c_en) {
    char data_read[NCV768X_ID_READALL_R_SIZE];
    if(!readI2C(NCV768X_ID_READALL_R, NCV768X_ID_READALL_R_SIZE, data_read)) {
        return false; // I2C NACK
    }
    data_read[0] = data_read[0] & 0b01111111;
    char data_write[NCV768X_ID_I2C_CONF_W_SIZE] = {(data_read[0] | (i2c_en << 7)), data_read[1]};
    return writeI2C(NCV768X_ID_I2C_CONF_W, NCV768X_ID_I2C_CONF_W_SIZE, data_write);
}

/**
 * @brief Changes diagnostic mode to OFF, AUTO_RETRY, or DIAG_ONLY.
 * @param diag_mode Enum to select the diagnostic mode.
 * OFF = No open load detection is performed.
 * AUTO_RETRY = During open load fault the following occurs. If fault is recovered, DIAG is released and normal operation continues.
 *      1) DIAG pin is pulled low.
 *      2) Low current is imposed on faulty channel only.
 *      3) Other channels turned off.
 * DIAG_ONLY = During open load fault, the DIAG pin is pulled low with no change to current regulation. If fault is recovered, DIAG is released.
 */
bool Ncv768x::i2cDiagMode(DiagMode diag_mode) {
    char data_read[NCV768X_ID_READALL_R_SIZE];
    if(!readI2C(NCV768X_ID_READALL_R, NCV768X_ID_READALL_R_SIZE, data_read)) {
        return false; // I2C NACK
    }
    data_read[0] = data_read[0] & 0b10011111;
    if (diag_mode == DiagMode::OFF) {
        // Do nothing, set already set to zero
    } else if (diag_mode == DiagMode::AUTO_RETRY) {
        data_read[0] = data_read[0] | 0b01000000;
    } else if (diag_mode == DiagMode::DIAG_ONLY) {
        data_read[0] = data_read[0] | 0b00100000;
    } else {
        return false;
    }
    char data_write[NCV768X_ID_I2C_CONF_W_SIZE] = {data_read[0] , data_read[1]};
    return writeI2C(NCV768X_ID_I2C_CONF_W, NCV768X_ID_I2C_CONF_W_SIZE, data_write);
}

/**
 * @brief Enables or disables the PWM dimming feature of OUT1-12.
 * @param pwm_en Boolean that enables or disables the PWM dimming enable for all OUT1-12, true = PWM enabled and I2C_CONF ignored, false = no PWM.
 */
bool Ncv768x::pwmEnable(bool pwm_en) {
    char data_read[NCV768X_ID_READALL_R_SIZE];
    if(!readI2C(NCV768X_ID_READALL_R, NCV768X_ID_READALL_R_SIZE, data_read)) {
        return false; // I2C NACK
    }
    data_read[0] = data_read[0] & 0b11101111;
    char data_write[NCV768X_ID_I2C_CONF_W_SIZE] = {data_read[0] | (pwm_en << 4), data_read[1]};
    return writeI2C(NCV768X_ID_I2C_CONF_W, NCV768X_ID_I2C_CONF_W_SIZE, data_write);    
}

/**
 * @brief Configures the the enable status of OUT1-12 in non-PWM mode.
 * @param i2c_ch_enable Boolean array containing the enable status of OUT1-12. When in PWM mode this enable status will always read enabled when in PWM mode.
 */
bool Ncv768x::i2cChEnableWrite(bool i2c_ch_enable[ch_size]) {
    char data_read[NCV768X_ID_READALL_R_SIZE];
    if(!readI2C(NCV768X_ID_READALL_R, NCV768X_ID_READALL_R_SIZE, data_read)) {
        return false; // I2C NACK
    }
    char16_t value = ((data_read[0] << 8) | data_read[1]) & 0b1111000000000000;
    for (uint8_t i = 0; i < ch_size; i++) {
        value = value | (i2c_ch_enable[i] << i);
    }
    
    char data_write[NCV768X_ID_I2C_CONF_W_SIZE] = {(value >> 8), value & 0xFF};
    return writeI2C(NCV768X_ID_I2C_CONF_W, NCV768X_ID_I2C_CONF_W_SIZE, data_write);
}

/**
 * @brief Configures all PWM registers. Returns false when pwm_freq or pwm_duty are out of range.
 * @param pwm_freq Enum that sets the PWM frequency to 125, 250, or 500 Hz for the NCV7684 or 150, 300, 600, or 1200 Hz for NCV7685. 
 * @param pwm_lin Boolean that sets PWM gain to linear or logarithmic, true = linear, false = logarithmic.
 * @param pwm_duty Integer that sets the PWM (duty cycle) value from 0 to 127 for each OUT channel. The NCV7684 does not have individual PWM duty so pwm_duty[0] is used for all channels.
 * @param pwm_en Boolean array that enables or disables PWM functionality of OUT1-12. 
 */
bool Ncv768x::pwmConf(PwmFreq pwm_freq, bool pwm_lin, uint8_t pwm_duty[ch_size], bool pwm_en[ch_size]) {
    // NCV7684 is was configured on initialization 
    if (m_part_number == PartNumber::NCV7684) {
        char data_write_pwm_conf[NCV768X_ID_PWM_CONF_NCV7684_W_SIZE] = {0b00000000};
        char data_write_pwm[NCV768X_ID_PWM_NCV7684_W_SIZE] = {0b00000000};
        // pwm_freq
        if (pwm_freq == PwmFreq::HZ_500) {
            data_write_pwm_conf[0] = data_write_pwm_conf[0] | (pwm_lin << 1) | 0b1;
        } else if ((pwm_freq == PwmFreq::HZ_250) | (pwm_freq == PwmFreq::HZ_125)) {
            data_write_pwm_conf[0] = data_write_pwm_conf[0] | (pwm_lin << 1) | 0b0;

        } else {
            return false;
        }
        // pwm_duty, take first array value
        if (pwm_duty[0] <= 127) {
            data_write_pwm[0] = data_write_pwm[0] | pwm_duty[0]; // don't care about PWM_GAIN bit 7 (PWMF1) as PWMF2 is set to 1
            if (pwm_freq == PwmFreq::HZ_250) {
                data_write_pwm[0] = data_write_pwm[0] | 0b10000000;
            } else if (pwm_freq == PwmFreq::HZ_125) {
                // do nothing already zero from definition
            }
        }
        else {
            return false; // invalid gain setting
        }
        // pwm_en
        data_write_pwm[1] = (pwm_en[11] << 3) | (pwm_en[10] << 2) | (pwm_en[9] << 1) | (pwm_en[8]);
        data_write_pwm[2] = (pwm_en[7] << 7) | (pwm_en[6] << 6) | (pwm_en[5] << 5) | (pwm_en[4] << 4) |
                            (pwm_en[3] << 3) | (pwm_en[2] << 2) | (pwm_en[1] << 1) | (pwm_en[0]);
        
        return (writeI2C(NCV768X_ID_PWM_NCV7684_W, NCV768X_ID_PWM_NCV7684_W_SIZE , data_write_pwm) 
                && writeI2C(NCV768X_ID_PWM_CONF_NCV7684_W, NCV768X_ID_PWM_CONF_NCV7684_W_SIZE, data_write_pwm_conf));

    // NCV7685 is was configured on initialization 
    } else if (m_part_number == PartNumber::NCV7685) {
        char data_write_pwm_conf[NCV768X_ID_PWM_CONF_NCV7685_W_SIZE] = {0b00000000};
        // pwm_lin
        data_write_pwm_conf[0] = data_write_pwm_conf[0] | (pwm_lin << 2);
        // pwm_freq
        if (pwm_freq == PwmFreq::HZ_150) {
            // do nothing already zero from definition
        } else if (pwm_freq == PwmFreq::HZ_300) {
            data_write_pwm_conf[0] = data_write_pwm_conf[0] | 0b00000001;
        } else if (pwm_freq == PwmFreq::HZ_600) {
            data_write_pwm_conf[0] = data_write_pwm_conf[0] | 0b00000010;
        } else if (pwm_freq == PwmFreq::HZ_1200) {
            data_write_pwm_conf[0] = data_write_pwm_conf[0] | 0b00000011;
        } else {
            return false;
        }
        // pwm_en
        data_write_pwm_conf[1] = (pwm_en[11] << 3) | (pwm_en[10] << 2) | (pwm_en[9] << 1) | (pwm_en[8]);
        data_write_pwm_conf[2] = (pwm_en[7] << 7) | (pwm_en[6] << 6) | (pwm_en[5] << 5) | (pwm_en[4] << 4) |
                                 (pwm_en[3] << 3) | (pwm_en[2] << 2) | (pwm_en[1] << 1) | (pwm_en[0]);
        // pwm_duty
        char data_write_pwm_all[NCV768X_ID_PWM_ALL_NCV7685_W_SIZE] = {0b00000000};
        for (uint8_t i = 0; i < ch_size; i++) {
            data_write_pwm_all[i] = pwm_duty[i];
        }
        return (writeI2C(NCV768X_ID_PWM_CONF_NCV7685_W, NCV768X_ID_PWM_CONF_NCV7685_W_SIZE , data_write_pwm_conf) 
                && writeI2C(NCV768X_ID_PWM_ALL_NCV7685_W, NCV768X_ID_PWM_ALL_NCV7685_W_SIZE, data_write_pwm_all));
    } else {
        return false;
    }
    
}

// ================ SAM and OTP Registers ================ //

/**
 * @brief Check to see if part has been OTP'ed. First call checkI2C() to ensure I2C communication. 
 * This function will return false (not OTP'ed or OTP'ed with default values) or true (OTP'ed).
 * Note: It is not possible to determine if part was OTP'ed if it was OTP'ed with all default register values.
 */
bool Ncv768x::checkOtp() {

    // Read OTP registers to check them against default values
    uint8_t id_vers_1;
    uint8_t id_vers_2;
    bool sam_conf_1[ch_size];
    bool sam_conf_2[ch_size];
    DiagMode diag;
    bool crc;
    uint8_t addr;
    readOtp(&id_vers_1, &id_vers_2, sam_conf_1, sam_conf_2, &diag, &crc, &addr);
    
    // Check sam_conf registers for differences from default values
    // All sam_conf_1 bits are true by default for NCV7684 and false for NCV7685
    bool sam_conf_1_value;
    if (m_part_number == PartNumber::NCV7684) {
        sam_conf_1_value = true;
    } else if (m_part_number == PartNumber::NCV7685) {
        sam_conf_1_value = false;
    }
    for (uint8_t i = 0; i < ch_size; i++) {
        if ((sam_conf_1[i] == !sam_conf_1_value) || (sam_conf_2[i] == false)) {
            return true; // Has been OTP'ed
        }
    }
    // Check diag register for difference between default values
    if (diag != DiagMode::DIAG_ONLY) {
        return true; // Has been OTP'ed
    }

    // Cannot check crc or addr because:
    // I2C communication would fail if OTP'ed with crc or addr that are different than default values
    // Use checkI2C() function to determine valid I2C communication

    return false; // Not OTP'ed or OTP values are default
}


/**
 * @brief Reads the entire NCV768X_ID_READ_OTP register. 
 * For the NCV7684, CSN must be above 1.66V otherwise this function will fail and return false.
 * For the NCV7685, ISET must be between 2.5V and 3.3V otherwise this function will fail and return false.
 * @param id_vers_1 Pointer to byte that contains NCV768X_ID_VERS_1 register.
 * @param id_vers_2 Pointer to byte that contains NCV768X_ID_VERS_2 register.
 * @param sam_conf_1 Boolean array of OUT1-12 enable states for SAM mode 1.
 * @param sam_conf_2 Boolean array of OUT1-12 enable states for SAM mode 2.
 * @param diag_mode Enum to select the diagnostic mode.
 * OFF = No open load detection is performed.
 * AUTO_RETRY = During open load fault the following occurs. If fault is recovered, DIAG is released and normal operation continues.
 *      1) DIAG pin is pulled low.
 *      2) Low current is imposed on faulty channel only.
 *      3) Other channels turned off.
 * DIAG_ONLY = During open load fault, the DIAG pin is pulled low with no change to current regulation. If fault is recovered, DIAG is released.
 * @param crc_en Pointer to boolean that indicates if I2C CRC mode is enabled, true = CRC enabled, false = CRC disabled.
 * @param bus_addr Pointer to a byte that contains the current I2C 7-bit slave address.
 */
bool Ncv768x::readOtp(uint8_t* id_vers_1, uint8_t* id_vers_2, bool sam_conf_1[ch_size], bool sam_conf_2[ch_size], DiagMode* diag_mode, bool* crc_en, uint8_t* bus_addr) {
    char data[NCV768X_ID_READ_OTP_R_SIZE];
    if(!readI2C(NCV768X_ID_READ_OTP_R, NCV768X_ID_READ_OTP_R_SIZE, data)) {
        return false; // I2C NACK
    }

    *id_vers_1 = data[0];
    *id_vers_2 = data[1];

    char16_t sam_conf_1_char16 = (data[2] << 8) | data[3];
    char16_t sam_conf_2_char16 = (data[4] << 8) | data[5];
    for (uint8_t i = 0; i < ch_size; i++) {
        sam_conf_1[i] = (sam_conf_1_char16 >> i) & 0b1;
        sam_conf_2[i] = (sam_conf_2_char16 >> i) & 0b1;
    }

    bool autor = (data[6] >> 7) & 0b1;
    bool detonly = (data[6] >> 6) & 0b1;
    if (autor && detonly) {
        // Case where both autor and detonly are true should not occur on read
        return false;
    } else if ((autor == false) && (detonly == false)) {
        // No diagnostic 
        *diag_mode = DiagMode::OFF;
    } else if (autor) {
        *diag_mode = DiagMode::AUTO_RETRY;
    } else if (detonly) {
        *diag_mode = DiagMode::DIAG_ONLY;
    } else {
        return false;
    }

    *crc_en = m_crc_en;

    *bus_addr = m_i2c_address >> 1;

    return true;
}

/**
 * @brief Writes the either the NCV768X_ID_SET_OTP or NCV768X_ID_LOCK_OTP register. 
 * For the NCV7684, CSN must be above 1.66V otherwise this function will fail and return false.
 * For the NCV7685, ISET must be between 2.5V and 3.3V otherwise this function will fail and return false.
 * @param otp Set to true to OTP the device and false to write only to NCV768X_ID_SET_OTP registers. WARNING, OTP can be done once!
 * @param sam_conf_1 Boolean array of OUT1-12 enable states for SAM mode 1.
 * @param sam_conf_2 Boolean array of OUT1-12 enable states for SAM mode 2.
 * @param diag_mode Enum to select the diagnostic mode.
 * OFF = No open load detection is performed.
 * AUTO_RETRY = During open load fault the following occurs. If fault is recovered, DIAG is released and normal operation continues.
 *      1) DIAG pin is pulled low.
 *      2) Low current is imposed on faulty channel only.
 *      3) Other channels turned off.
 * DIAG_ONLY = During open load fault, the DIAG pin is pulled low with no change to current regulation. If fault is recovered, DIAG is released.
 * @param crc_en Boolean that indicates if I2C CRC mode is enabled, true = CRC enabled, false = CRC disabled.
 * @param bus_addr Byte that sets a new I2C 7-bit slave address. Valid range is from 0x60 to 0x7F.
 */
bool Ncv768x::writeOtp(bool otp, bool sam_conf_1[ch_size], bool sam_conf_2[ch_size], DiagMode diag_mode, bool crc_en, uint8_t bus_addr) {
    char data_write[NCV768X_ID_OTP_W_SIZE] = {0b00000000};

    // sam_conf_1
    data_write[0] = (sam_conf_1[11] << 3) | (sam_conf_1[10] << 2) | (sam_conf_1[9] << 1) | (sam_conf_1[8]);
    data_write[1] = (sam_conf_1[7] << 7) | (sam_conf_1[6] << 6) | (sam_conf_1[5] << 5) | (sam_conf_1[4] << 4) |
                    (sam_conf_1[3] << 3) | (sam_conf_1[2] << 2) | (sam_conf_1[1] << 1) | (sam_conf_1[0]);
    // sam_conf_2
    data_write[2] = (sam_conf_2[11] << 3) | (sam_conf_2[10] << 2) | (sam_conf_2[9] << 1) | (sam_conf_2[8]);
    data_write[3] = (sam_conf_2[7] << 7) | (sam_conf_2[6] << 6) | (sam_conf_2[5] << 5) | (sam_conf_2[4] << 4) |
                    (sam_conf_2[3] << 3) | (sam_conf_2[2] << 2) | (sam_conf_2[1] << 1) | (sam_conf_2[0]);

    // sam_add_set
    if (diag_mode == DiagMode::OFF) {
        // do nothing already zero from definition
    } else if (diag_mode == DiagMode::DIAG_ONLY) {
        data_write[4] = data_write[4] | 0b01000000;
    } else if (diag_mode == DiagMode::AUTO_RETRY) {
        data_write[4] = data_write[4] | 0b10000000;
    } else {
        return false;
    }

    // crc_en
    data_write[4] = data_write[4] | (crc_en << 5);

    // bus_addr
    if ((bus_addr >= 0x60) && (bus_addr <= 0x7F)) {
        data_write[4] = data_write[4] | (bus_addr & 0b00011111);
    } else {
        // Invalid I2C slave address. Valid range is from 0x60 to 0x7F.
        return false;
    }

    bool success = false;
    if (otp) {
        success = writeI2C(NCV768X_ID_LOCK_OTP_W, NCV768X_ID_OTP_W_SIZE, data_write);
    } else {
        success = writeI2C(NCV768X_ID_SET_OTP_W, NCV768X_ID_OTP_W_SIZE, data_write);
    }
    
    if (success) {
        // Only change CRC state and I2C address if I2C write was successful
        m_crc_en = crc_en; // Update CRC to be performed on subsequent set/lock of OTP registers
        m_i2c_address = bus_addr << 1; // 7-bit shifted left 1 to get 8-bit address
    }

    return success;
}
